
import boto3
from botocore.exceptions import ClientError

session = boto3.Session()

ec2_client = session.client('ec2')
ec2_resource = session.resource('ec2')

iam_client = session.client('iam')
iam_resource = session.resource('iam')

s3_client = session.client('s3')
s3_resource = session.resource('s3')


def get_aws_account_id():
    """ Returns AWS account ID"""
    response = ec2_client.describe_security_groups(GroupNames=['Default'])
    return response['SecurityGroups'][0]['OwnerId']


def create_bucket(region, bucket_name):
    """ Creates bucket to store application packages
    :param region:                  Region to create bucket in
    :param bucket_name:             Name of bucket
    :return:                        Name of bucket
    """
    try:
        s3_resource.create_bucket(Bucket=bucket_name
                                  , CreateBucketConfiguration={'LocationConstraint': region})
    except ClientError as exception:
        if exception.response['Error']['Code'] == 'BucketAlreadyOwnedByYou':
            print('Bucket already exists and owned by you, continuing')
        else:
            raise exception
    else:
        print('Created bucket: ' + bucket_name)

    return bucket_name


def upload_to_s3(bucket_name, file_object, file_name, fingerprint):
    """ Uploads file to S3
    :param bucket_name:                 Bucket name
    :param file_object:                 File to upload
    :param file_name:                   Name of zip file in S3
    :param fingerprint:                 Fingerprint of file_object
    """
    bucket = s3_resource.Bucket(bucket_name)

    try:
        cur_fingerprint = bucket.Object(file_name).metadata.get('fingerprint')
    except ClientError as e:
        if e.response['Error']['Message'] == 'Not Found':
            print('No package deployed yet. Uploading.')
        else:
            raise e
    else:
        if cur_fingerprint == fingerprint:
            print('Local package matches deployed. Not uploading.')
            return
        else:
            print('Local package does not match deployed. Uploading')

    bucket.put_object(Body=file_object, Key=file_name, Metadata={'fingerprint': fingerprint})


def get_subnet(subnet_id):
    return ec2_resource.Subnet(subnet_id)


def create_security_group(group_name, project_id, vpc_id, *rules):
    try:
        group = ec2_resource.create_security_group(
            Description='Bokchoi default security group',
            GroupName=group_name,
            VpcId=vpc_id
        )
    except ClientError as e:
        if e.response['Error']['Code'] == 'InvalidGroup.Duplicate':
            print('Security group already exists ' + group_name)
            return
        else:
            raise e

    group.create_tags(Tags=[{'Key': 'bokchoi-id', 'Value': project_id}])
    for rule in rules:
        group.authorize_ingress(**rule)
    print('Created security group: ' + group_name)
    return group


def get_security_groups(project_id, *group_names):

    filters = [{'Name': 'tag-key',
                'Values': ['bokchoi-id']},
               {'Name': 'tag-value',
                'Values': [project_id]}]

    if group_names:
        filters.append({'Name': 'group-name', 'Values': [*group_names]})

    response = ec2_client.describe_security_groups(Filters=filters)

    return [ec2_resource.SecurityGroup(group['GroupId']) for group in response['SecurityGroups']]


def delete_security_group(group, dryrun=True):

    if dryrun:
        print('Dryrun flag set. Would have deleted security group ' + group.group_name)
        return

    group_name = group.group_name
    group.delete()

    print('Deleted security group ' + group_name)


def create_instance_profile(profile_name, role_name=None):
    """ Creates IAM instance profile
    :param profile_name:            Name of profile to be created
    :param role_name:               Name of role to attach to instance profile
    :return:                        API response
    """
    try:
        create_instance_profile_response = iam_client.create_instance_profile(
            InstanceProfileName=profile_name
        )
    except ClientError as e:
        if e.response['Error']['Code'] == 'EntityAlreadyExists':
            print('Instance profile already exists ' + profile_name)
            return
        else:
            raise e

    if role_name:
        iam_client.add_role_to_instance_profile(
            InstanceProfileName=profile_name,
            RoleName=role_name
        )
    print('Created instance profile: ' + profile_name)
    return create_instance_profile_response['InstanceProfile']


def create_policy(policy_name, document):
    """ Creates IAM policy
    :param policy_name:             Name of policy to create
    :param document:                Policy document associated with policy
    """
    try:
        iam_client.create_policy(PolicyName=policy_name
                                 , PolicyDocument=document)
        print('Created policy: ' + policy_name)
    except ClientError as e:
        if e.response['Error']['Code'] == 'EntityAlreadyExists':
            print('Policy already exists ' + policy_name)
        else:
            raise e


def create_role(role_name, trust_policy, *policies):
    """ Creates IAM role
    :param role_name:               Name of role to create
    :param trust_policy:            Trust policy to associate with role
    :param policies:                Policies to attach to role
    :return:                        API response
    """
    try:
        iam_client.create_role(RoleName=role_name
                               , AssumeRolePolicyDocument=trust_policy)
    except ClientError as e:
        if e.response['Error']['Code'] == 'EntityAlreadyExists':
            print('Role already exists ' + role_name)
            return iam_resource.Role(role_name)
        else:
            raise e

    for policy in policies:
        if not policy:
            continue
        iam_client.attach_role_policy(
            RoleName=role_name,
            PolicyArn=policy.arn
        )
    print('Created role: ' + role_name)
    return iam_resource.Role(role_name)


def request_spot_instances(project_id, launch_spec, spot_price):
    """ Create spot instance request
    :param project_id:                  Global project id
    :param launch_spec:                 EC2 launch specification
    :param spot_price:                  Max price to bid for spot instance
    """
    response = ec2_client.request_spot_instances(LaunchSpecification=launch_spec
                                                 , SpotPrice=spot_price)

    spot_request_id = response['SpotInstanceRequests'][0]['SpotInstanceRequestId']

    waiter = ec2_client.get_waiter('spot_instance_request_fulfilled')
    waiter.wait(SpotInstanceRequestIds=[spot_request_id])

    ec2_client.create_tags(Resources=[spot_request_id]
                           , Tags=[{'Key': 'bokchoi-id', 'Value': project_id}])

    response = ec2_client.describe_spot_instance_requests(SpotInstanceRequestIds=[spot_request_id])
    instance_ids = [request['InstanceId'] for request in response['SpotInstanceRequests']]

    ec2_client.create_tags(Resources=instance_ids, Tags=[{'Key': 'bokchoi-id', 'Value': project_id}])


def cancel_spot_request(project_id, dryrun):
    """ Cancels spot instance request. Request is found by filtering on project_id tag.
    :param project_id:              Global project id
    :param dryrun:                  If true list id's of spot requests to cancel
    """
    print('\nCancelling spot request')
    filters = [{'Name': 'tag:bokchoi-id', 'Values': [str(project_id)]}
               , {'Name': 'state', 'Values': ['open', 'active']}]
    response = ec2_client.describe_spot_instance_requests(Filters=filters)

    spot_request_ids = [request['SpotInstanceRequestId'] for request in response['SpotInstanceRequests']]

    if dryrun:
        print('Dryrun flag set. Would have cancelled spot requests ' + ', '.join(spot_request_ids))
        return

    try:
        ec2_client.cancel_spot_instance_requests(SpotInstanceRequestIds=spot_request_ids)
        print('Spot requests cancelled')
    except ClientError as e:
        if e.response['Error']['Code'] == 'InvalidParameterCombination':
            print('No spot requests to cancel')
        else:
            raise e


def get_instances(project_id):
    """ Returns all instances for project. Instances are found by filtering on project_id tag
    :param project_id:
    :return:
    """
    filters = [{'Name': 'tag:bokchoi-id', 'Values': [str(project_id)]},
               {'Name': 'instance-state-name', 'Values': ['pending', 'running', 'stopping', 'stopped']}]
    return list(ec2_resource.instances.filter(Filters=filters))


def terminate_instance(instance, dryrun=True):
    """ Terminates instance.
    :param instance:                ec2.Instance
    :param dryrun:                  If True print instance that would be terminated
    """
    print('\nTerminating instances')

    if dryrun:
        print('Dryrun flag set. Would have terminated instance: ' + instance.instance_id)
        return

    instance.terminate()
    instance.wait_until_terminated()
    print('Instance terminated')


def delete_bucket(project_id, dryrun=True):
    """ Delete Bokchoi deploy bucket. Removes all object it contains.
    :param project_id:              Global project id
    :param dryrun:                  If True list bucket that would be terminated
    """
    print('\nDelete Bucket')

    bucket = s3_resource.Bucket(project_id)

    if dryrun:
        print('Dryrun flag set. Would have deleted bucket ' + bucket.name)
        return

    try:
        bucket.objects.delete()
        bucket.delete()
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchBucket':
            pass
        else:
            raise e


def get_instance_profiles(project_id):
    """ Yields all instance profiles associated with deployment
    :param project_id:              Global project id
    """
    for instance_profile in iam_resource.instance_profiles.all():
        if project_id in instance_profile.instance_profile_name:
            yield instance_profile


def delete_instance_profile(instance_profile, dryrun):
    """ Deletes instance profile. First removes all roles attached to instance profile.
    :param instance_profile:        Name of instance profile
    :param: dryrun:                 If true print name of instance profile to delete
    """
    instance_profile_name = instance_profile.instance_profile_name
    print('\nDeleting Instance Profile:', instance_profile_name)

    if dryrun:
        print('Dryrun flag set. Would have deleted instance profile ' + instance_profile_name)
        return

    try:
        for role in instance_profile.roles_attribute:
            instance_profile.remove_role(RoleName=role['RoleName'])
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            pass
        else:
            raise e

    try:
        instance_profile.delete()
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            pass
        else:
            raise e

    print('Successfully deleted Instance Profile:', instance_profile_name)


def get_roles(project_id):
    """ Yields all IAM roles associated with deployment
    :param project_id:              Global project id
    :return:                        IAM role
    """
    for role in iam_resource.roles.all():
        if project_id in role.role_name:
            yield role


def delete_role(role, dryrun):
    """ Deletes IAM role. First detaches all polices from role
    :param role:                    Boto3 Role resource
    :param dryrun:                  If true print name of role to delete
    """
    role_name = role.role_name
    print('\nDeleting Role:', role_name)

    if dryrun:
        print('Dryrun flag set. Would have deleted role ' + role_name)
        return

    try:
        for policy in role.attached_policies.all():
            policy.detach_role(RoleName=role_name)
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            print('No policies to detach')
        else:
            raise e

    try:
        role.delete()
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            print('Role does not exist')
        else:
            raise e

    print('Successfully deleted role:', role_name)


def get_policies(project_id, pattern=None):
    """ Returns all IAM policies associated with deployment
    :param project_id:              Global project id
    :param pattern:                 Pattern to return specific policies (e.g. default-policy)
    :return:                        Boto3 policy resource
    """

    policies = []

    for policy in iam_resource.policies.filter(Scope='Local'):

        policy_name = policy.policy_name

        if project_id not in policy_name:
            continue

        if pattern and pattern not in policy_name:
            continue

        policies.append(policy)

    return policies


def delete_policy(policy, dryrun):
    """ Deletes IAM policy. First detaches all roles from policy.
    :param policy:                  Boto3 policy resource
    :param dryrun:                  If true print name of policy to delete
    """
    policy_name = policy.policy_name
    print('\nDeleting Policy:', policy_name)

    if dryrun:
        print('Dryrun flag set. Would have deleted policy ' + policy_name)
        return

    try:
        for role in policy.attached_roles.all():
            policy.detach_role(RoleName=role.role_name)
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            print('Role for policy does not exist')
        else:
            raise e

    try:
        policy.delete()
    except ClientError as e:
        if e.response['Error']['Code'] == 'NoSuchEntity':
            print('Policy does not exist')
        else:
            raise e

    print('Successfully deleted Policy:', policy_name)



#!/usr/bin/env python3.6
"""
Class which can be used to deploy and run EMR jobs
"""

import os
import sys
import time

import boto3

from bokchoi import utils
from bokchoi.aws import common

class EMR(object):
    """Create EMR object which can be used to schedule jobs"""
    def __init__(self, project, settings):
        self.settings = settings
        self.project_name = project

        aws_account_id = common.get_aws_account_id()
        self.project_id = utils.create_project_id(project, aws_account_id)
        self.job_flow_id = None

    def deploy(self, path=''):
        """Zip package and deploy to S3 so it can be used by EMR"""
        bucket_name = common.create_bucket(self.settings['Region'], self.project_id)

        cwd = os.getcwd()
        package, fingerprint = utils.zip_package(path or cwd, self.settings.get('Requirements'))

        package_name = 'bokchoi-' + self.project_name + '.zip'
        common.upload_to_s3(bucket_name, package, package_name, fingerprint)

    def run(self):
        """Create Spark cluster and run specified job"""
        emr_client = boto3.client('emr')
        self.start_spark_cluster(emr_client)
        self.step_prepare_env(emr_client)
        self.step_spark_submit(emr_client)

    def undeploy(self, dryrun):
        """Deletes all policies, users, and instances permanently"""

        common.cancel_spot_request(self.project_id, dryrun)
        common.terminate_instances(self.project_id, dryrun)

        for pol in common.get_policies(self.project_id):
            common.delete_policy(pol, dryrun)

        for prof in common.get_instance_profiles(self.project_id):
            common.delete_instance_profile(prof, dryrun)

        for role in common.get_roles(self.project_id):
            common.delete_role(role, dryrun)

        # remove s3 bucket
        common.delete_bucket(self.project_id, dryrun)

    def start_spark_cluster(self, emr_client):
        """
        Start Spark cluster based on configuration given in settings
        """

        launch_spec = self.settings['EMR']['LaunchSpecification']

        instance_type = launch_spec['InstanceType']
        instance_count = self.settings['EMR']['InstanceCount']

        instances = {'KeepJobFlowAliveWhenNoSteps': False
                     , 'TerminationProtected': False
                     , 'Ec2SubnetId': self.settings['EMR']['LaunchSpecification']['SubnetId']
                     , 'InstanceGroups': [
                            {
                                'Name': 'EmrMaster',
                                'Market': 'SPOT',
                                'InstanceRole': 'MASTER',
                                'BidPrice': self.settings['EMR']['SpotPrice'],
                                'InstanceType': instance_type,
                                'InstanceCount': 1
                            },
                            {
                                'Name': 'EmrCore',
                                'Market': 'SPOT',
                                'InstanceRole': 'CORE',
                                'BidPrice': self.settings['EMR']['SpotPrice'],
                                'InstanceType': instance_type,
                                'InstanceCount': instance_count - 1
                            }
                        ]
                     }

        additional_sgs = launch_spec.get('AdditionalSecurityGroups')
        if additional_sgs:
            instances['AdditionalMasterSecurityGroups'] = additional_sgs
            instances['AdditionalSlaveSecurityGroups'] = additional_sgs

        response = emr_client.run_job_flow(
            Name=self.project_id,
            LogUri="s3://{}/spark/".format(self.project_id),
            ReleaseLabel=self.settings['EMR']['Version'],
            Instances=instances,
            Configurations=[
                {"Classification": "spark-env"
                 , "Properties": {}
                 , "Configurations": [
                        {
                            "Classification": "export",
                            "Properties": {
                                "PYSPARK_PYTHON": "python34"
                            },
                            "Configurations": []
                        }
                    ]
                }
            ],
            Tags=[{'Key': 'bokchoi-id', 'Value': self.project_id}],
            Applications=[{'Name': 'Hadoop'}, {'Name': 'Spark'}],
            JobFlowRole='EMR_EC2_DefaultRole',
            ServiceRole='EMR_DefaultRole',
            VisibleToAllUsers=True,
        )
        # parse EMR response to check if successful
        response_code = response['ResponseMetadata']['HTTPStatusCode']
        if response['ResponseMetadata']['HTTPStatusCode'] == 200:
            self.job_flow_id = response['JobFlowId']
        else:
            print("Error creating: (status code {})".format(response_code))
            sys.exit(1)

        print("Created Spark cluster with job {}".format(self.job_flow_id))

    def step_prepare_env(self, emr_client):
        """Copies files from S3 and unzips them"""
        package_name = 'bokchoi-' + self.project_name + '.zip'
        s3_package_uri = 's3://{bucket}/{key}'.format(bucket=self.project_id, key=package_name)
        root_dir = '/home/hadoop/'

        emr_client.add_job_flow_steps(
            JobFlowId=self.job_flow_id,
            Steps=[
                {
                    'Name': 'setup - copy files',
                    'ActionOnFailure': 'CANCEL_AND_WAIT',
                    'HadoopJarStep': {
                        'Jar': 'command-runner.jar',
                        'Args': ['aws', 's3', 'cp', s3_package_uri, root_dir]
                    }
                },
                {
                    'Name': 'setup - unzip files',
                    'ActionOnFailure': 'CANCEL_AND_WAIT',
                    'HadoopJarStep': {
                        'Jar': 'command-runner.jar',
                        'Args': ['unzip', root_dir + package_name, '-d', root_dir]
                    }
                },
                {
                    'Name': 'setup - install python dependencies',
                    'ActionOnFailure': 'CANCEL_AND_WAIT',
                    'HadoopJarStep': {
                        'Jar': 'command-runner.jar',
                        'Args': ['sudo', 'pip-3.4', 'install', '-r', root_dir + 'requirements.txt']
                    }
                }
            ]
        )

    def step_spark_submit(self, emr_client):
        """Submit spark job given by user"""
        py3_env = 'PYSPARK_PYTHON=/usr/bin/python3'
        script = self.settings['EntryPoint']

        emr_client.add_job_flow_steps(
            JobFlowId=self.job_flow_id,
            Steps=[
                {
                    'Name': 'Run Spark',
                    'ActionOnFailure': 'CANCEL_AND_WAIT',
                    'HadoopJarStep': {
                        'Jar': 'command-runner.jar',
                        'Args': ['spark-submit', '--conf', py3_env, '/home/hadoop/' + script]
                    }
                }
            ]
        )
        print("Added step 'spark-submit'")
        time.sleep(1)  # Prevent ThrottlingException



"""
Can be used to run Bokchoi on the Google Cloud using Google Compute Engines

"""

import os
import sys
import time
import bokchoi.utils

import googleapiclient.discovery
import googleapiclient.errors

from google.auth import exceptions as auth_except, default
from google.oauth2 import service_account
from google.cloud import storage, exceptions


class GCP(object):
    """Run Bokchoi on the Google Cloud using Google Compute Engines"""
    def __init__(self, bokchoi_project_name, settings):
        self.project_name = bokchoi_project_name
        self.entry_point = settings['EntryPoint']
        self.requirements = settings.get('Requirements', [])
        self.wait_for_execution = settings.get("WaitForExecution", False)
        self.gcp = self.retrieve_gcp_settings(settings)
        self.credentials = self.authorize_client()
        self.compute = self.get_authorized_compute()
        self.storage = self.get_authorized_storage()

    def authorize_client(self):
        """If the environment variable GOOGLE_APPLICATION_CREDENTIALS or If the Google Cloud SDK is
        installed and has application default credentials set they are loaded and returned."""
        try:
            credentials, _ = default()
            return credentials
        except auth_except.DefaultCredentialsError as e:
            credentials = service_account.Credentials.from_service_account_file(self.gcp.get('auth_key'))
            return credentials
        except Exception as e:
            print('Authentication failed, please set the GOOGLE_APPLICATION_CREDENTIALS env variable, install the'
                  'Google SDK and authenticate, or supply the JSON file location. \n', e)

    def get_authorized_compute(self):
        """Authorize with default method (implicit env variable) or otherwise
        use the explicit authentication.
        """
        return googleapiclient.discovery.build('compute', 'v1', credentials=self.credentials)

    def get_authorized_storage(self):
        """Authorize with default method (implicit env variable) or otherwise
        use the explicit authentication.
        """
        return storage.Client(credentials=self.credentials)

    @staticmethod
    def retrieve_gcp_settings(settings):
        """
        Checks the given settings and validates if all required values
        are there and the values are correct.
        :arg settings: a json file with with defined settings
        :return: a python dict with renamed input parameters + defaults
        """
        gcp = settings['GCP']

        def check_none(v):
            if not gcp.get(v):
                raise Exception('{} is required, please add it to the config'.format(v))

        required = ['ProjectId', 'AuthKeyLocation', 'Bucket']
        [check_none(v) for v in required]

        return {
            'project': gcp.get('ProjectId'),
            'auth_key': gcp.get('AuthKeyLocation'),
            'bucket': gcp.get('Bucket'),
            'region': gcp.get('Region', 'europe-west4'),
            'zone': gcp.get('Zone', 'europe-west4-b'),
            'network': gcp.get('Network', 'default'),
            'sub_network': gcp.get('SubNetwork', 'default'),
            'instance_type': gcp.get('InstanceType', 'n1-standard-1'),
            'preemptible': gcp.get('Preemptible', False),
            'disk_space': gcp.get('DiskSpaceGb', 25)
        }

    def list_instances(self):
        """List names of all existing instances"""
        result = self.compute.instances().list(
            project=self.gcp.get('project'),
            zone=self.gcp.get('zone')).execute()
        instances = [x['name'] for x in result['items']]
        return instances

    def define_instance_config(self):
        """
        Set up a compute engine configuration based on the user's input
        :return: Defined Compute Engine configuration
        """
        image_response = self.compute.images().getFromFamily(
            project='debian-cloud', family='debian-9').execute()

        machine_type = "zones/{}/machineTypes/{}".format(
            self.gcp.get('zone'), self.gcp.get('instance_type'))

        gcp_script = os.path.join(os.path.dirname(__file__), 'gcp-startup-script.sh')
        with open(gcp_script, 'r') as script:
            startup_script = script.read()

        config = {
            'name': self.project_name,
            'machineType': machine_type,

            'disks': [
                {
                    'boot': True,
                    'autoDelete': True,
                    'initializeParams': {
                        'sourceImage': image_response['selfLink'],
                        'diskSizeGb': self.gcp.get('disk_space')
                    }
                }
            ],
            'scheduling': {
              'preemptible': self.gcp.get('preemptible')
            },

            'networkInterfaces': [{
                'network': 'global/networks/{}'.format(self.gcp.get('network')),
                'subnetwork': '/regions/{}/subnetworks/{}'.format(
                    self.gcp.get('region'), self.gcp.get('sub_network')),
                'accessConfigs': [
                    {'type': 'ONE_TO_ONE_NAT', 'name': 'External NAT'}
                ]
            }],

            'serviceAccounts': [{
                'email': 'default',
                'scopes': [
                    'https://www.googleapis.com/auth/devstorage.read_write',
                    'https://www.googleapis.com/auth/logging.write',
                    'https://www.googleapis.com/auth/compute'
                ]
            }],

            'metadata': {
                'items': [{
                    'key': 'startup-script',
                    'value': startup_script
                }, {
                    'key': 'bucket_name',
                    'value': self.gcp.get('bucket')
                }, {
                    'key': 'package_name',
                    'value': '{}-{}.zip'.format(self.project_name, 'package')
                }, {
                    'key': 'entry_point',
                    'value': self.entry_point
                }, {
                    'key': 'instance_name',
                    'value': self.project_name
                }, {
                    'key': 'zone',
                    'value': self.gcp.get('zone')
                }]
            }
        }

        return config

    def create_instance(self):
        """Create a new compute engine"""
        print('Creating instance')
        try:
            return self.compute.instances().insert(
                project=self.gcp.get('project'),
                zone=self.gcp.get('zone'),
                body=self.define_instance_config()).execute()
        except googleapiclient.errors.HttpError as e:
            if 'already exists' in str(e):
                print('instance with name {} already exists. exit(1)'.format(self.project_name))
                sys.exit(1)
            else:
                print(e)
                sys.exit(1)

    def delete_instance(self):
        """Remove the created compute engine"""
        print('Deleting instance')
        return self.compute.instances().delete(
            project=self.gcp.get('project'),
            zone=self.gcp.get('zone'),
            instance=self.project_name).execute()

    def wait_for_operation(self, operation):
        """Method which polls the status of the operations and returns when the
        operation is completed.
        :arg operation: a gcp api operation
        """
        if operation is None:
            return

        print('Waiting for operation to finish...')
        while True:
            result = self.compute.zoneOperations().get(
                project=self.gcp.get('project'),
                zone=self.gcp.get('zone'),
                operation=operation['name']).execute()

            if result['status'] == 'DONE':
                if 'error' in result:
                    raise Exception(result['error'])
                return result

            time.sleep(3)

    def create_bucket(self):
        """Create a new storage bucket which will be used for the defined job"""
        print('Creating bucket')
        try:
            bucket = self.storage.create_bucket(self.gcp.get('bucket'))
            return bucket
        except exceptions.Conflict as e:
            if 'You already own this bucket' in str(e):
                print('Bucket with name {} already exists, skipping create.'.format(self.gcp.get('bucket')))
            else:
                print(e)

    def delete_bucket(self):
        """Delete the created bucket"""
        print('Deleting bucket')
        try:
            bucket = self.storage.get_bucket(self.gcp.get('bucket'))
            bucket.delete(force=True)
        except exceptions.NotFound as e:
            print('bucket does not exist, skipping deletion')

    def upload_blob(self, file_name, file_object):
        """Upload file to Google Storage
        :arg file_name: target filename in Google storage
        :arg file_object: zip file object which will be uploaded
        :return: public url of the Google Storage resource
        """
        bucket = self.storage.get_bucket(self.gcp.get('bucket'))
        blob = bucket.blob(file_name)
        blob.upload_from_file(file_object)
        return blob.public_url

    def download_blob(self, file_name):
        """Download file from Google Storage
                :arg file_name: target filename in Google storage
                :return: fileobject as string
                """
        bucket = self.storage.get_bucket(self.gcp.get('bucket'))
        blob = bucket.blob(file_name)
        filestring = blob.download_as_string().decode('utf-8')
        return filestring

    def deploy(self, path):
        """Deploy package to GCP/Google Storage"""
        print('Uploading package to Google Storage bucket')
        self.create_bucket()
        package, fingerprint = bokchoi.utils.zip_package(path, self.requirements)
        self.upload_blob('{}-{}.zip'.format(self.project_name, 'package'), package)
        return 'Deployed!'

    def undeploy(self, dryrun=False):
        """Undeploy and delete all created resources"""
        print('Deleting resources which are created on GCP')
        self.delete_bucket()
        return 'Undeployed!'

    def run(self):
        """Run the uploaded package"""
        create_instance_op = self.create_instance()
        self.wait_for_operation(create_instance_op)
        print('Running application')
        if not self.wait_for_execution:
            return "Not awaiting execution, finishing now"
        sec = 0
        while True:
            if self.project_name in self.list_instances():
                time.sleep(10)
                sec += 10
                if sec % 300 == 0:
                    print("minute {}: instance still running".format(sec / 60))
            else:
                logs = self.download_blob('{}-{}.zip-logs.txt'.format(self.project_name, 'package'))
                print("""Instance no longer running, logs are as follows:\n\n-------------------------------------
                --------------------------------------------""")
                for l in logs.split('\n'):
                    print(l)
                print("---------------------------------------------------------------------------------")
                break

        return 'Script has finished'

    def stop(self, dryrun=False):
        return 'Stop not yet implemented. Please stop VM manually'

    def connect(self, dryrun, *args, **kwargs):
        print('Connect not yet implemented')

    def status(self):
        print('Status not yet implemented')
